// (C) 2003 Ulrich Hecht <uli@emulinks.de>
// if2a patch by D. Gauchard <deyv@free.fr>
// Licensed under the terms of the GNU Public License version 2

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "f2a.h"
#include <unistd.h>
#include <stdlib.h>

#include "f2aloader.gba-src.h"

//////////////////////////////
// for f2a 256Mbits
const int cartsize_mb = 256;
int romblocksize_bit = 15;		// 15=32KB, has to be 17(128KB) for Turbo carts ?
int writeblocksize_bit = 18;	// 256KB - empiric
//////////////////////////////

#define GBA_HEADNAME	"GBAROM-"
#define GBA_EXTNAME	".gba"

#define MODE_WRITE	0
#define MODE_READ	1
#define MODE_UNDEF	2

#define MAX(x,y)	((x)>(y)? (x): (y))
#define MIN(x,y)	((x)<(y)? (x): (y))
#define ASCII(x)	(isascii(x)? (x): '.')


#define RAMBASE		0xe000000
#define	ROMBASE		0x8000000

#define CARTSIZE_mb 	(cartsize_mb)
#define CARTSIZE_B 	(CARTSIZE_mb * 1024 * 1024 / 8)
#define LOADERSIZE_B	sizeof(f2aloader_gba)

int romblocksize = -1;
int writeblocksize = -1;
int loader = 1;
int verbose = 0;
int force = 0;
int map = 0;
int sram = 0;
int thorough = 0;
int reduce_forced = 0;
int reduce_allowed = 1;
int reduce = 0;
int max_rom_size = 0;

void help (char* name)
{
	fprintf(stderr,
		"if2a " F2A_VERSION "\n"
		"Usage: %s options [file [file [...]]]\n"
		"Options are:\n"
		"	-M	display rom map\n"
		"	-R	read rom (and generate filenames)\n"
		"	-S <s>	maximum cart size (suffix kb,mb,kB,mB, default %imb)\n"
		"	-r	read sram\n"
		"	-W	write rom\n"
		"	-w	write sram\n"
		"	-b <b>	specify block in sram (**)\n"
		"	-s	do not try to reduce rom size (default: reduce to fit loader)\n"
		"	-a	always try to reduce rom size (default: reduce to fit loader)\n"
		"	-d	dummy - do not write -\n"
		"	-d	(one more -d) - do not read -\n"
		"	-v	be more verbose\n"
		"debug only:\n"
		"	-c	(with -W) check whole file (not just borders) before burning\n"
		"	-n	do not insert f2a loader (*)\n"
		"\nNotes:\n"
		". Concerning rom files: you should always specify whole cart's content on cmdline,\n"
		"  cart will be updated according to differences between cmdline and cart's content\n"
		". *  Normally, f2a's loader is automatically inserted in beginning of rom\n"
		". ** A block can be '1' or '2a' or '3b2' (same format as f2a's rom loader)\n"
		"	1a1 (8KB) \\\n"
		"	1a2 (8KB)  \\  1a   \\\n"
		"	1a3 (8KB)  / (32KB) \\\n"
		"	1a4 (8KB) /          \\   1\n"
		"	1b1 (8KB) \\          / (64KB)\n"
		"	1b2 (8KB)  \\  1b    /\n"
		"	1b3 (8KB)  / (32KB)/\n"
		"	1b4 (8KB) /\n"
		"	2a1 (8KB)\n"
		"	...\n"
		"	4b4 (8KB)\n",
		name,
		CARTSIZE_mb);
}

void setup ()
{
	assert(sizeof(sendmsg)==63);
	assert(sizeof(recvmsg)==64);
	romblocksize = 1<<romblocksize_bit;
	writeblocksize = 1<<writeblocksize_bit;
	max_rom_size = CARTSIZE_B;
}

int has_good_header (unsigned char* rom)
{
	return memcmp(f2aloader_gba, rom, 0xa0)==0? 1: -1;
}

void correct_header (unsigned char* rom, unsigned char* name)
{
	unsigned char romname[13];
	unsigned char romcode[5];
	memset(romname, 32, 12);
	memset(romcode, 32, 12);
	romname[12] = romcode[4] = 0;
	memcpy(romname, name, MIN(strlen(name), 12));
	memcpy(romcode, name, MIN(strlen(name), 4));
	
	if (!isalnum(rom[0xa0]))
		memcpy(&rom[0xa0], romname, 12);
	if (!isalnum(rom[0xac]))
		memcpy(&rom[0xac], romcode, 4);
	//if (!isalnum(rom[0xb0]))
	//	rom[0xb0] = '0';
	
	if (has_good_header(rom) < 0)
	{
		fprintf(stderr, "Warning: no standard header found in '%s', rewriting it\n", romname);
		memcpy(rom, f2aloader_gba, 0xa0);
	}
}

void display_map (unsigned char* rom)
{
	int i;
	

	printf("Game Title:	");
	for (i = 0xa0; i < 0xac; i++)
		printf("%c", ASCII(rom[i]));
	printf("\nGame Code:	AGB-");
	for (i = 0xac; i < 0xb0; i++)
		printf("%c", ASCII(rom[i]));
	printf("\nMaker Code:	");
	for (i = 0xb0; i < 0xb1; i++)
		printf("%c", ASCII(rom[i]));
	printf("\n");
	//printf("\n0x96:		0x%02x\n", rom[0xb2]);
}

unsigned char* romname (unsigned char* rom)
{
	int len;
	static char name [13];
	
	for (len = 0; len < 12; len++)
		name[len] = ASCII(rom[0xa0 + len]);
	for (len--; len>=0 && (name[len]==32 || name[len]=='.'); len--);
	name[len+1] = 0;
	return name;
}

void display_cart_map (void)
{
	int i;
	unsigned char rom[LOADUNIT];
	
	if (sim > 1)
		return;

	for (i = ROMBASE; i < ROMBASE + max_rom_size; i += romblocksize)
	{
		printf("Trying address 0x%x...\r", i);
		f2a_readmem(rom, i, LOADUNIT);
		if (has_good_header(rom) > -1)
		{
			printf("\n");
			display_map(rom);
		}
	}
}

void auto_readandsave_rom (int numfiles, char** files)
{
	int i;
	int indexfile;
	int nextblockloaded;
	char* name;
	char autoname[64];
	unsigned char rom[romblocksize];
	
	assert(romblocksize >= LOADUNIT);
	
	if (sim > 1)
		return;

	indexfile = 0;
	i = ROMBASE;
	f2a_readmem(rom, i, LOADUNIT);
	do
	{
		nextblockloaded = 0;
		
		printf("Trying address 0x%x...\r", i);
		if (has_good_header(rom) > -1)
		{
			name = romname(rom);
			strcpy(autoname, GBA_HEADNAME);
			strcat(autoname, name);
			strcat(autoname, GBA_EXTNAME);
			
			printf("\n");
			if (verbose)
				printf("Detected rom '%s'\n", name);
			
			if (indexfile >= numfiles)
				name = autoname;
			else
				name = files[indexfile];
			indexfile++;
			
			if (strcmp(name, "@") == 0)
				name = autoname;
			
			if (strcmp(name, ".") == 0) // user don't want this rom
			{
				if (verbose)
					printf("Skipping on user request\n");
			}
			else
			{
				int f;
				if (verbose)
					printf("Using name file name '%s'\n", name);
				if ((f = open(name, O_WRONLY|O_CREAT|O_EXCL,0666)) == -1)
					perror(name);
				else
				{
					f2a_readmem(rom, i, romblocksize);
					do
					{
						if (write(f, rom, romblocksize) == -1)
						{
							perror("write");
							break;
						}
						if ((i+=romblocksize) < ROMBASE + max_rom_size)
							f2a_readmem(rom, i, romblocksize);
					} while (i < ROMBASE + max_rom_size && has_good_header(rom) == -1);
					close(f);
					nextblockloaded = 1;
				}
			}
		}
		
		if (!nextblockloaded)
		{
			if ((i+=romblocksize) < ROMBASE + max_rom_size)
				f2a_readmem(rom, i, LOADUNIT);
		}
	} while (i < ROMBASE + max_rom_size);
	printf("\n");
}

void display_memory_map (unsigned char* rom, int size)
{
	int i;

	for (i = 0; i < size; i += romblocksize)
	{
		printf("Trying address 0x%x...\r", i);
		if (has_good_header(&rom[i]) > -1)
		{
			printf("\n");
			display_map(&rom[i]);
		}
	}
}

int is_same (int offset, int size, unsigned char* mem)
{
	int ret, i;
	unsigned char rom [LOADUNIT];
	
	if (sim > 1)
		return -1;

	for (i = offset; i < offset + size; i += LOADUNIT)
	{
		// check beginning
		f2a_readmem(rom, ROMBASE + i, LOADUNIT);
		ret = memcmp(rom, &mem[i], LOADUNIT);
		if (verbose > 1 && ret != 0)
			printf("(offset 0x%x-0x%x BAD)\n", i, i + LOADUNIT - 1);
		if (ret != 0)
			return -1;
	}
	return 1;
}

int has_same_data (int offset, int size, unsigned char* mem)
{
	if (thorough)
	{
		int i;
		for (i = offset; i < offset + size; i += romblocksize)
			if (is_same(i, romblocksize, mem) == -1)
				return -1;
	}
	else
	{
		// check beginnning
		if (is_same(offset, LOADUNIT, mem) == -1)
			return -1;

		// roughly check end
		size &= ~(LOADUNIT-1);
		if (is_same(offset + size - LOADUNIT, LOADUNIT, mem) == -1)
			return -1;
	}
	return 1;
}

int conv (char* number)
{
	char* unit = number;
	while (isdigit(*unit))
		unit++;
	if (strlen(unit) == 2)
	{
		int byte;
		switch (tolower(unit[0]))
		{
		case 'k':	byte = 1<<10; break;
		case 'm':	byte = 1<<20; break;
		default:	byte = -8;
		}
		switch (unit[1])
		{
		case 'b':	byte >>= 3; break;
		case 'B':	break;
		default:	byte = -1;
		}
		if (byte > 0)
		{
			*unit = 0;
			return atoi(number) * byte;
		}
	}
	fprintf(stderr, "number '%s' not recognized\n", number);
	exit(1);
	return -1;
}

int convf2ablock (char* block, int* offset, int* size )
{
	int len = strlen(block);

	if (len > 3)
		return -1;

	*offset = (block[0] - '1') << 16;
	*size = 1 << 16;
	if (block[1])
	{
		*offset += (tolower(block[1]) - 'a') << 15;
		*size = 1 << 15;
		
		if (block[2])
		{
			*offset += (block[2] - '1') << 13;
			*size = 1 << 13;
		}
	}
	if (verbose)
		printf("Block %s is SRAM is (offset=0x%x/%i size=0x%x/%i)\n",
			block, *offset, *offset, *size, *size);
	return 0;
}

void adjust_burn_addresses (int* offset, int* size)
{
	// adjust offset so that it's a multiple of block
	// adjust size so that it's a multiple of block
	// (this is needed if we want to rewrite a rom between two others)
	*size += *offset;
	*offset &= ~(writeblocksize - 1);
	*size -= *offset;
	if (*size & (writeblocksize - 1))
		*size += writeblocksize;
	*size &= ~(writeblocksize - 1);
}

void adjust_whole_size (int* size)
{
	if (*size & (writeblocksize - 1))
		*size += writeblocksize;
	*size &= ~(writeblocksize - 1);
}

void adjust_cart_size (int* size)
{
	if (*size & (romblocksize - 1))
		*size += romblocksize;
	*size &= ~(romblocksize - 1);
}

//////////////////////////////////////////////////////////////////////////
// check whole size

#define STAT							\
		if (j == -1)					\
			st.st_size = sizeof(f2aloader_gba);	\
		else if (stat(files[j], &st) == -1)		\
		{						\
			perror(files[j]);			\
			return -1;				\
		} 

int get_wholesize (int numfiles, char** files)
{
	int j;
	struct stat st;
	int wholesize = 0;
	int roundedfilesize;
	
	// stat every files
	for (j = loader? -1: 0; j < numfiles; j++)
	{
		if(verbose) printf("Stat file: %s\n", j==-1? "Loader": files[j]);

		STAT

		roundedfilesize = st.st_size;
		adjust_cart_size(&roundedfilesize);
		wholesize += roundedfilesize;
	}
	
/*
	if (wholesize > max_rom_size + sizeof(f2aloader_gba))
	{
		fprintf(stderr, "You have not enough memory !\n");
		exit(1);
	}
*/
	if (wholesize > max_rom_size)
		// We hope that we will be able to reduce one of the rom's size by sizeof(f2aloader_gba)
		wholesize = max_rom_size;
	
	adjust_whole_size(&wholesize);
	return wholesize;
}

//////////////////////////////////////////////////////////////////////////
// this one is simple, it writes one file to sram

int loadandwrite_sram (char* file, int offset, int size)
{
	int sramf = -1;
	struct stat st;
	char* sram = NULL;
	
	if (stat(file, &st)==-1)
	{
		perror(file);
		return -1;
	}

	// some checks
	if (st.st_size > size)
	{
		fprintf(stderr, "File too big (0x%x bytes) to fit in 0x%x bytes\n", (int)st.st_size, size);
		return -1;
	}

	// allocate memory

	if ((sram = (unsigned char*)malloc(size)) == NULL)
	{
		fprintf(stderr, "Cannot allocate 0x%x bytes - ", size);
		perror("malloc");
		return -1;
	}
	memset(sram, 0xff, size);

	// load files into "allocated rom"
	if (verbose)
		printf("Loading file: %s (size=0x%x)\n", file, (int)st.st_size);

	if ((sramf = open(file, O_RDONLY)) == -1)
		perror("open");
	if (read(sramf, sram, size) == -1)
	{
		if (verbose > 1)
			fprintf(stderr, "while reading 0x%x bytes: ", size);
		perror("failed to read from file");
		return -1;
	}
	close(sramf);
	f2a_writemem(sram, RAMBASE, offset, size, size);

	return 0;
}

//////////////////////////////////////////////////////////////////////////
// automatic check cart and burn if necessary

void burnrom (char* rom, int offset, int size)
{
	printf("\nBurning from 0x%x to 0x%x", offset, offset + size);
	adjust_burn_addresses(&offset, &size);
	printf(" -> 0x%x to 0x%x\n", offset, offset + size);
	f2a_writemem(&rom[offset], ROMBASE, offset, size, writeblocksize);
}

int auto_loadandburn_rom (int numfiles, char** files)
{
	int j;
	int romf = -1;
	struct stat st;
	int loadedsize;
	int roundedfilesize;
	char* rom;
	int burnstart = -1;
	int wholesize = get_wholesize(numfiles, files);

	if (!wholesize) return -1;
	
	// allocate memory
	if ((rom = (unsigned char*)malloc(wholesize)) == NULL)
	{
		fprintf(stderr, "Cannot allocate 0x%x bytes - ", wholesize);
		perror("malloc");
		return -1;
	}
	memset(rom, 0xff, wholesize);
	
	// load
	loadedsize = 0;
	for (j = loader? -1: 0; j < numfiles; j++)
	{
		printf("\n");
		printf("File %s at offset 0x%x is:\n", j==-1? "Loader": files[j], loadedsize);
		
		STAT
		
		roundedfilesize = st.st_size;
		adjust_cart_size(&roundedfilesize);

		if (j == -1)
		{
			// copy f2aloader
			assert(loadedsize == 0);
			memcpy(rom, f2aloader_gba, sizeof(f2aloader_gba));
		}
		else
		{
			if (loadedsize + st.st_size > max_rom_size)
			{
				fprintf(stderr, "! WARNING ! Truncating file (size 0x%x -> 0x%x bytes) !\n", (int)st.st_size, max_rom_size - loadedsize);
				if ((st.st_size = max_rom_size - loadedsize) == 0)
					break;
				roundedfilesize = st.st_size;
				adjust_cart_size(&roundedfilesize);
			}

			// load file
			if ((romf = open(files[j], O_RDONLY)) == -1)
				perror("open");
			if (read(romf, &rom[loadedsize], st.st_size) == -1)
			{
				if (verbose > 1)
					fprintf(stderr, "while reading 0x%x bytes: ", (int)st.st_size);
				perror("failed to read from file");
				return -1;
			}
			close(romf);
			
			// try to reduce rom size
			if (reduce_allowed || reduce_forced)
			{
				int i;
				char last = rom[loadedsize + st.st_size - 1];
				for (i = st.st_size - 2; i >= 0 && rom[loadedsize + i] == last; i--);
				roundedfilesize = i + 2;
				adjust_cart_size(&roundedfilesize);
				if (roundedfilesize < st.st_size)
				{
					printf("(reducing rom size from %ikB to %ikB", ((int)st.st_size + 1023) >> 10, (roundedfilesize + 1023) >> 10);
					if ((reduce += ((int)st.st_size - roundedfilesize)) > sizeof(f2aloader_gba))
					{
						printf(" - not trying to further reduce rom size");
						reduce_allowed = 0;
					}
					printf(")\n");
				}
				st.st_size = roundedfilesize;
			}
		}

		// check file against real rom
		correct_header(&rom[loadedsize], j==-1? "CARTBOOT": files[j]);
		display_map(&rom[loadedsize]);
		if (!force && has_same_data(loadedsize, st.st_size, rom) >= 0)
		{
			printf("No need to burn it!\n");
			// but it's time to burn the previous ones if they changed
			if (burnstart >= 0)
			{
				burnrom(rom, burnstart, loadedsize - burnstart);
				burnstart = -1;
			}
		}
		else
		{
			printf("Not the same !\n");
			// and we remember start address
			if (burnstart < 0)
				burnstart = loadedsize;
		}

		loadedsize += roundedfilesize;
	}
	if (burnstart >= 0)
		burnrom(rom, burnstart, loadedsize - burnstart);

	return 0;
}

//////////////////////////////////////////////////////////////////////////
// main

int main(int argc, char** argv)
{
	recvmsg rm;
	int opt;
	int mode = MODE_UNDEF;
	int map = 0;
	char* block = NULL;
	int size = -1, offset = -1;

	setup();

	if (argc == 1)
	{
		help(argv[0]);
		exit(5);
	}

	do
	{
		opt=getopt(argc, argv, "hMRrWwb:nfdvcsaS:");
		switch(opt)
		{
		case 'r':
			sram = 1;
		case 'R':
			mode = MODE_READ;
			break;
			
		case 'w':
			sram = 1;
		case 'W':
			mode = MODE_WRITE;
			break;
			
		case 'b':
			block = optarg;
			break;

		case 'v':
			verbose++;
			break;
		
		case 'n':
			loader = 0;
			break;
		
		case 'f':
			force = 1;
			break;
		
		case 'c':
			thorough = 1;
			break;
		
		case 's':
			reduce_allowed = 0;
			break;
		
		case 'a':
			reduce_forced = 1;
			break;
		
		case 'S':
			max_rom_size = conv(optarg);
			printf("Setting rom size to %iMbits / %iKBytes\n",
				max_rom_size >> 17,
				max_rom_size >> 10);
			break;
			
		case 'd':
			sim++;
			break;
		
		case 'M':
			map = 1;
			break;

		case 'h':
		case '?':
			help(argv[0]);
			exit(1);
		}
	} while(opt!=-1);

	if (mode == MODE_UNDEF && !map)
	{
		fprintf(stderr, "No action specified\n");
		exit(1);
	}
	
	if (argc-optind <= 0 && (mode == MODE_WRITE || sram))
	{
		fprintf(stderr, "File name missing\n");
		exit(1);
	}

	if (sram && mode == MODE_READ && argc-optind != 1)
	{
		fprintf(stderr, "Cannot write SRAM to multiple files\n");
		exit(1);
	}

	if (sim < 2 && usb_connect())
	{
		fprintf(stderr, "Unable to connect to F2A linker\n");
		exit(1);
	}
	
	if (block)
	{
		if (convf2ablock(block, &offset, &size) < 0)
		{
			fprintf(stderr, "Block %s has bad format\n", block);
			exit(1);
		}
	}

	memset(&rm,0,sizeof(rm));
	if (!(sim > 1))
	{
		f2a_info(&rm);
		if (rm.data[0] != 4)
		//while (rm.data[0] != 4)
		{
			fprintf(stderr, "Please turn OFF then ON your GBA with SELECT and START held down.\n");
			f2a_boot();
			f2a_info(&rm);
			if (verbose)
				printf("Firmware uploaded.\n");
		}
	}

	if (mode == MODE_UNDEF && map)
	{
		display_cart_map();
		exit(0);
	}
	
	if (sram)
	{
		if (mode == MODE_WRITE)
		{
			if (verbose)
				printf("Uploading to SRAM offset 0x%x -> 0x%x to file %s\n", RAMBASE + offset, RAMBASE + offset + size - 1, argv[optind]);
			loadandwrite_sram(argv[optind], offset, size);
		}
		else
		{
			if (verbose)
				printf("Downloading from SRAM offset 0x%x -> 0x%x to file %s\n", RAMBASE + offset, RAMBASE + offset + size - 1, argv[optind]);
			f2a_readmemtofile(argv[optind], RAMBASE + offset, size);
		}
	}
	else
	{
		if (mode == MODE_WRITE)
			auto_loadandburn_rom(argc-optind, argv+optind);
		else
			auto_readandsave_rom(argc-optind, argv+optind);
	}

	if (sim < 2)
		usb_disconnect();

	return 0;
}
